# This file was automatically generated from a file in the source DreamPie dir.
# DO NOT EDIT IT, as your changes will be gone when the file is created again.

__version__ = "1.2"
release_timestamp = 1348531200 # calendar.timegm((2012, 9, 25, 0, 0, 0))
